@@ -15,7 +15,7 @@ import logging
 
 # reqman imports
 from . import pycode
-from .common import assert_syntax
+from .common import assert_syntax, RqException
 from . import tool
 from .ehttp import MyHeaders
 
@@ -90,29 +90,30 @@ def jzon_dumps(o,indent:int|None=2):
 
 class Env:
     def __init__(self, /, **kwargs):
-        self._data:dict = _convert(kwargs)
-        self._compile_py_methods()
+        self._data={}
         self.__params_scopes: list = []
+        self.update( kwargs )
 
     def __setitem__(self, key, value):
-        self._data[key] = _convert(value)
-        self._compile_py_methods()
+        self.update( {key:value} )
 
     def update(self, dico):
-        ###################################################
-        # self._data.update(_convert(dico))
-        # self._compile_py_methods()
-        ###################################################
-        # to be able to set vars at switch time
-        for k,v in dico.items():
-            if isinstance(v,str) and ( v.startswith("<<") or v.startswith("{{")) and ( v.endswith(">>") or v.endswith("}}")):
-                self.__setitem__(k,self.substitute(v))
-            else:
-                self.__setitem__(k,v)
-
+        self._data.update(_convert(dico))
+        self._compile_py_methods()
 
     def __getitem__(self, key):
-        return self._data[key]
+        v=self._data[key]
+        if isinstance(v,str) and ("<<" in v or "{{" in v):
+            try:
+                new_v = self.substitute(v)
+                if new_v is not v:
+                    # if substitution happened, cache the result for eager-like behavior
+                    self._data[key] = new_v
+                return new_v
+            except:
+                 pass # if substitution fails, return original value
+        return v
+
 
     def __contains__(self, key):
         return key in self._data
@@ -137,8 +138,12 @@ class Env:
         if code in os.environ:
             return os.environ[code]
 
-        env = dict(self._data)
-        result = eval(code, {}, env)
+        try:
+            env = {k:self[k] for k in self._data}
+            result = eval(code, dict(tool=tool), env)
+        except Exception as e:
+            raise RqException(f"Error evaluating expression '{code}': {e}") from e
+
         if with_context:
             try:
                 vars_in_expr = {node.id for node in ast.walk(ast.parse(code)) if isinstance(node, ast.Name)}
@@ -162,7 +167,8 @@ class Env:
                 try:
                     val = self.eval(expr)
                 except Exception as e:
-                    val = f"***ERROR: {e}***"
+                    #val = f"***ERROR: {e}***"
+                    val = l # return the original <<expr>>
             logger.debug(f"SUBSTITUTE {l} by {val} ({type(val)})")
             if isinstance(val, str):
                 text = text.replace(l, val)
@@ -175,10 +181,10 @@ class Env:
                     text = text.replace(l, jzon_dumps(val, indent=None))
         return text
 
-    def substitute_in_object(self, o: Any) -> Any:
+    def substitute_in_object(self, o: Any, raise_error: bool = True) -> Any:
         def _sub_in_object(o: Any) -> Any:
             if isinstance(o, str):
-                return self.substitute(o)
+                return self.substitute(o,raise_error)
             elif isinstance(o, dict):
                 return {k: _sub_in_object(v) for k, v in o.items()}
             elif isinstance(o, list):
@@ -236,6 +242,7 @@ class Env:
                 for k, v in d.items():
                     code = pycode.is_python(k, v)
                     if code:
+                        logger.warning(f"Security warning: Compiling and executing python method '{k}'. Ensure that the code is from a trusted source.")
                         scope = {}
                         exec(code, dict(ENV=self, tool=tool), scope)  # declare ENV&tool in method!
                         d[k] = scope[k]
@@ -246,6 +253,9 @@ class Env:
                     declare_methods(d[i])
         declare_methods(self._data)
 
+    def __repr__(self):
+        return jzon_dumps(self._data)
+
 if __name__ == "__main__":
     ...
     # logging.basicConfig(level=logging.DEBUG)
